/**
 * Current state of a player's color grading effect.
 */
struct ColorGradeState {

  // LUT currently visible and being transitioned to.
  int fromLut;
  int toLut;
  double currentSpeed;

  // Once the current transition has finished, this will be the next LUT transitioned to.
  int queuedLut;
  double queuedSpeed;

  // How far along the transition the state is.
  double alpha;
}


/**
 * Initializes color grading state for players.
 */
class ColorGradeEventHandler : EventHandler {
  override void PlayerEntered(PlayerEvent e) {
    let thinker = ColorGradeThinker.Get();
    thinker.InitPlayer(e.PlayerNumber);
  }

  override void PlayerRespawned(PlayerEvent e) {
    let thinker = ColorGradeThinker.Get();
    thinker.InitPlayer(e.PlayerNumber);
  }
}


/**
 * Main thinker for initializing and updating the color grading shader.
 */
class ColorGradeThinker : Thinker {

  // Default transition speed.
  const DEFAULT_SPEED_TICKS = 70;

  // State for all possible players.
  ColorGradeState playerStates[MAXPLAYERS];

  // Was color grading enabled on the previous tick?
  bool wasEnabled;

  /**
   * Thinker initialization.
   */
  ColorGradeThinker Init() {
    self.wasEnabled = bool(CVar.FindCVar('r_colorgrading').GetInt());

    return self;
  }

  /**
   * Returns the only thinker instance.
   */
  static ColorGradeThinker Get() {
    ThinkerIterator it = ThinkerIterator.Create("ColorGradeThinker");
    let p = ColorGradeThinker(it.Next());
    if (p == null) {
      p = new("ColorGradeThinker").Init();
    }
    return p;
  }

  override void Tick() {
    bool enabled = bool(CVar.FindCVar('r_colorgrading').GetInt());
    if (!enabled && !wasEnabled) {
      return;
    }

    for (int i = 0; i < MAXPLAYERS; i++) {
      if (!playeringame[i]) {
        continue;
      }

      // Toggles the shader state from the current console variable state.
      if (!enabled && wasEnabled) {
        Shader.SetEnabled(players[i], "ColorGrade", false);
        continue;
      } else if (enabled && !wasEnabled) {
        Shader.SetEnabled(players[i], "ColorGrade", true);
      }

      // Transition is completed.
      if (playerStates[i].alpha >= 1.0) {

        // Start transitioning to the queued LUT.
        if (playerStates[i].toLut != playerStates[i].queuedLut) {
          playerStates[i].fromLut = playerStates[i].toLut;
          playerStates[i].toLut = playerStates[i].queuedLut;
          playerStates[i].currentSpeed = playerStates[i].queuedSpeed;
          playerStates[i].alpha = 0.0;
        }

      // Update current transition and adjust shader.
      } else {
        playerStates[i].alpha += playerStates[i].currentSpeed;
        if (playerStates[i].alpha >= 1.0) {
          playerStates[i].alpha = 1.0;
        }

      }

      Shader.SetUniform1i(players[i], "ColorGrade", "lutA", playerStates[i].fromLut);
      Shader.SetUniform1i(players[i], "ColorGrade", "lutB", playerStates[i].toLut);
      Shader.SetUniform1f(players[i], "ColorGrade", "alpha", playerStates[i].alpha);
    }

    self.wasEnabled = enabled;
  }

  /**
   * Initializes color grading for a single player.
   */
  void InitPlayer(int playerNumber) {
    self.PlayerSet(playerNumber, 0);
    Shader.SetEnabled(players[playerNumber], "ColorGrade", true);
  }

  /**
   * Sets up a transition from the current LUT to a new one.
   */
  void PlayerTransitionTo(int playerNumber, int newLut, int speed) {
    if (speed == 0) {
      speed = DEFAULT_SPEED_TICKS;
    }

    playerStates[playerNumber].queuedLut = newLut;
    playerStates[playerNumber].queuedSpeed = 1.0 / double(speed);
  }

  /**
   * Immediately sets a new LUT.
   */
  void PlayerSet(int playerNumber, int lut) {
    playerStates[playerNumber].fromLut = lut;
    playerStates[playerNumber].toLut = lut;
    playerStates[playerNumber].queuedLut = lut;
    playerStates[playerNumber].currentSpeed = 0.0;
    playerStates[playerNumber].queuedSpeed = 0.0;

    playerStates[playerNumber].alpha = 1.0;
  }

  /**
   * Transition to new LUT for a given player, callable from ACS.
   * If player is set to -1 all players will transition to the new LUT.
   */
  static void TransitionTo(int playerNumber, int lutNew, int speed) {
    let thinker = ColorGradeThinker.Get();

    if (playerNumber == -1) {
      for (int i = 0; i < MAXPLAYERS; i++) {
        if (!playeringame[i]) {
          continue;
        }
        thinker.PlayerTransitionTo(i, lutNew, speed);
      }

    } else {
      thinker.PlayerTransitionTo(playerNumber, lutNew, speed);

    }
  }

  /**
   * Set a new LUT for given player, callable from ACS.
   * If player isset to -1 all players will transition to the new LUT.
   */
  static void Set(int playerNumber, int lut) {
    let thinker = ColorGradeThinker.Get();

    if (playerNumber == -1) {
      for (int i = 0; i < MAXPLAYERS; i++) {
        if (!playeringame[i] ) {
          continue;
        }
        thinker.PlayerSet(i, lut);
      }

    } else {
      thinker.PlayerSet(playerNumber, lut);

    }
  }
}
